"""
Parse files and energies - wrap MDAnalysis functions.

The parsers module deals with input. This can be a trajectory or single-frame topology file (FileParser),
specifically a Pseudotrajectory file (PtParser) or energy write-out (XVGParser).

The result of file parsing is immediately transformed into a ParsedMolecule or a sequence of ParsedMolecule
objects -> ParsedTrajectory. Similarly, energies are stored in the ParsedEnergy object that can be connected to
ParsedTrajectory. Those are the objects that other modules should access.
"""

import os
from pathlib import Path
from typing import Generator, Tuple, List, Callable

import numpy as np

import pandas as pd
from MDAnalysis.auxiliary.XVG import XVGReader
from numpy.typing import NDArray
from scipy.spatial.transform import Rotation
import MDAnalysis as mda
from MDAnalysis.core import AtomGroup

from molgri.constants import EXTENSIONS_STR, EXTENSIONS_TRJ
from molgri.paths import PATH_INPUT_ENERGIES
from molgri.space.rotations import two_vectors2rot
from molgri.space.fullgrid import FullGrid


class ParsedMolecule:

    def __init__(self, atoms: AtomGroup, box=None):
        """
        Wraps the behaviour of AtomGroup and Box implemented by MDAnalysis and provides the necessary
        rotation and translation functions.

        Args:
            atoms: an AtomGroup that makes up the molecule
            box: the box in which the molecule is simulated
        """
        self.atoms = atoms
        self.box = box
        self._update_attributes()

    def _update_attributes(self):
        self.num_atoms = len(self.atoms)
        self.atom_labels = self.atoms.names
        self.atom_types = self.atoms.types

    def get_atoms(self) -> AtomGroup:
        return self.atoms

    def get_center_of_mass(self) -> NDArray:
        return self.atoms.center_of_mass()

    def get_positions(self) -> NDArray:
        return self.atoms.positions

    def get_box(self):
        return self.box

    def __str__(self):
        return f"<ParsedMolecule with atoms {self.atom_types} at {self.get_center_of_mass()}>"

    def _rotate(self, about_what: str, rotational_obj: Rotation, inverse: bool = False):
        if about_what == "origin":
            point = (0, 0, 0)
        elif about_what == "body":
            point = self.get_center_of_mass()
        else:
            raise ValueError(f"Rotation about {about_what} unknown, try 'origin' or 'body'.")
        if inverse:
            rotational_obj = rotational_obj.inv()
        self.atoms.rotate(rotational_obj.as_matrix(), point=point)

    def rotate_about_origin(self, rotational_obj: Rotation, inverse: bool = False):
        self._rotate(about_what="origin", rotational_obj=rotational_obj, inverse=inverse)

    def rotate_about_body(self, rotational_obj: Rotation, inverse: bool = False):
        self._rotate(about_what="body", rotational_obj=rotational_obj, inverse=inverse)

    def translate(self, vector: np.ndarray):
        self.atoms.translate(vector)

    def rotate_to(self, position: NDArray):
        """
        1) rotate around origin to get to a rotational position described by position
        2) rescale radially to original length of position vector

        Args:
            position: 3D coordinates of a point on a sphere, end position of COM
        """
        assert len(position) == 3, "Position must be a 3D location in space"
        com_radius = np.linalg.norm(self.get_center_of_mass())
        position_radius = np.linalg.norm(position)
        self.translate_radially(position_radius - com_radius)
        rot_matrix = two_vectors2rot(self.get_center_of_mass(), position)
        self.rotate_about_origin(Rotation.from_matrix(rot_matrix))


    def translate_to_origin(self):
        """
        Translates the object so that the COM is at origin. Does not change the orientation.
        """
        current_position = self.get_center_of_mass()
        self.translate(-current_position)

    def translate_radially(self, distance_change: float):
        """
        Moves the object away from the origin in radial direction for the amount specified by distance_change (or
        towards the origin if a negative distance_change is given). If the object is at origin, translate in z-direction
        of the internal coordinate system.

        Args:
            distance_change: the change in length of the vector origin-object
        """
        # need to work with rounding because gromacs files only have 3-point precision
        initial_vector = np.round(self.get_center_of_mass(), 3)
        if np.allclose(initial_vector, [0, 0, 0], atol=1e-3):
            initial_vector = np.array([0, 0, 1])
        len_initial = np.linalg.norm(initial_vector)
        rescaled_vector = distance_change*initial_vector/len_initial
        self.atoms.translate(rescaled_vector)


class FileParser:

    def __init__(self, path_topology: str, path_trajectory: str = None, path_energy: str = None):
        """
        This module serves to load topology or trajectory information and output it in a standard format of
        a ParsedMolecule or a generator of ParsedMolecules (one per frame). No properties should be accessed directly,
        all work in other modules should be based on attributes and methods of ParsedMolecule.

        Args:
            path_topology: a full path to the structure file (.gro, .xtc, .xyz ....) that should be read
            path_trajectory: a full path to the trajectory file (.xtc, .xyz ....) that should be read
        """
        self.path_topology = path_topology
        self.path_trajectory = path_trajectory
        self.path_energy = self._try_find_energy_file(path_energy)
        self.path_topology, self.path_trajectory = self._try_to_add_extension()
        if self.path_trajectory is not None and self.path_topology is not None:
            self.universe = mda.Universe(self.path_topology, self.path_trajectory)
        elif self.path_topology is not None:
            self.universe = mda.Universe(self.path_topology)
        elif self.path_trajectory is not None:
            self.universe = mda.Universe(self.path_trajectory)
        else:
            raise ValueError("FileParser cannot produce any output if path_topology and path_trajectory are None.")

    def _try_find_energy_file(self, path_energy):
        # file exists and is provided
        if path_energy is not None and os.path.isfile(path_energy):
            return path_energy
        else:
            # try based on other names
            name_getters = [self.get_topology_file_name(), self.get_trajectory_file_name()]
            for proposed_name in name_getters:
                if proposed_name is not None:
                    proposed_path = f"{PATH_INPUT_ENERGIES}{proposed_name}.xvg"
                    if os.path.isfile(proposed_path):
                        return proposed_path
        return None

    def _try_to_add_extension(self) -> List[str]:
        """
        If no extension is added, the program will try to add standard extensions. If no or multiple files with
        standard extensions are added, it will raise an error. It is always possible to explicitly state an extension.
        """
        possible_exts_structure = EXTENSIONS_STR
        possible_exts_traj = EXTENSIONS_TRJ
        possible_exts = (possible_exts_structure, possible_exts_traj)
        to_return = [self.path_topology, self.path_trajectory]
        for i, path in enumerate(to_return):
            if path is None:
                continue
            root, ext = os.path.splitext(path)
            if not ext:
                options = []
                for possible_ext in possible_exts[i]:
                    test_file = f"{path}.{possible_ext.lower()}"
                    if os.path.isfile(test_file):
                        options.append(test_file)
                # if no files found, complain
                if len(options) == 0:
                    raise FileNotFoundError(f"No file with name {path} could be found and adding standard "
                                            f"extensions did not help.")
                # if exactly one file found, use it
                elif len(options) == 1:
                    to_return[i] = options[0]
                # if several files with same name but different extensions found, complain
                else:
                    raise AttributeError(f"More than one file corresponds to the name {path}. "
                                         f"Possible choices: {options}. "
                                         f"Please specify the extension you want to use.")
        return to_return

    def get_dt(self):
        "Time in ps between trajectory frames (assuming constant time step)"
        return self.universe.trajectory[1].time - self.universe.trajectory[0].time

    def get_file_path_trajectory(self):
        return self.path_trajectory

    def get_file_path_topology(self):
        return self.path_topology

    def get_topology_file_name(self):
        if self.path_topology is None:
            return None
        return Path(self.path_topology).stem

    def get_trajectory_file_name(self):
        if self.path_trajectory is None:
            return None
        return Path(self.path_trajectory).stem

    def get_parsed_trajectory(self, default_atom_selection=None):
        if not self.path_energy:
            parsed_energies = None
        else:
            xvg_parser = XVGParser(self.path_energy)
            parsed_energies = xvg_parser.get_parsed_energy()

        name = self.get_topology_file_name()
        molecule_generator = self.generate_frame_as_molecule
        return ParsedTrajectory(name, molecule_generator, parsed_energies, is_pt=isinstance(self, PtParser),
                                default_atom_selection=default_atom_selection, dt=self.get_dt())

    def as_parsed_molecule(self) -> ParsedMolecule:
        """
        Method to use if you want to parse only a single molecule, eg. if only a topology file was provided.

        Returns:
            a parsed molecule from current or only frame
        """
        return ParsedMolecule(self.universe.atoms, box=self.universe.dimensions)

    def generate_frame_as_molecule(self, atom_selection = None) -> Generator[ParsedMolecule, None, None]:
        """
        Method to use if you want to parse an entire trajectory and generate a ParsedMolecule for each frame.

        Returns:
            a generator yielding each frame individually
        """
        if atom_selection is None:
            atom_selection = "all"
        for _ in self.universe.trajectory:
            yield ParsedMolecule(self.universe.select_atoms(atom_selection, sorted=False), box=self.universe.dimensions)


class PtParser(FileParser):

    def __init__(self, m1_path: str, m2_path: str, path_topology: str, path_trajectory: str = None,
                 path_energy: str = None):
        """
        A parser specifically for Pseudotrajectories written with the molgri.writers module. Useful to test the
        behaviour of writers or to analyse the resulting pseudo-trajectories.

        Args:
            m1_path: full path to the topology of the central molecule
            m2_path: full path to the topology of the rotating molecule
            path_topology: full path to the topology of the combined system
            path_trajectory: full path to the trajectory of the combined system
        """
        super().__init__(path_topology=path_topology, path_trajectory=path_trajectory, path_energy=path_energy)
        self.c_num = FileParser(m1_path).as_parsed_molecule().num_atoms
        self.r_num = FileParser(m2_path).as_parsed_molecule().num_atoms
        assert self.c_num + self.r_num == self.as_parsed_molecule().num_atoms

    def generate_frame_as_double_molecule(self) -> Generator[Tuple[ParsedMolecule, ParsedMolecule], None, None]:
        """
        Differently to FileParser, this generator always yields a pair of molecules: central, rotating (for each frame).

        Yields:
            (central molecule, rotating molecule) at each frame of PT

        """
        for _ in self.universe.trajectory:
            c_molecule = ParsedMolecule(self.universe.atoms[:self.c_num], box=self.universe.dimensions)
            r_molecule = ParsedMolecule(self.universe.atoms[self.c_num:], box=self.universe.dimensions)
            yield c_molecule, r_molecule

    def get_parsed_trajectory(self):
        pt = super().get_parsed_trajectory()
        pt.set_c_r(c_num=self.c_num, r_num=self.r_num)
        return pt

    def generate_r_molecule(self) -> Generator[ParsedMolecule, None, None]:
        return self.generate_frame_as_molecule(atom_selection=self.get_atom_selection_r())

    def generate_c_molecule(self) -> Generator[ParsedMolecule, None, None]:
        return self.generate_frame_as_molecule(atom_selection=self.get_atom_selection_c())


class ParsedEnergy:

    def __init__(self, energies: NDArray, labels, unit):
        self.energies = energies
        self.labels = labels
        self.unit = unit

    def get_energies(self, energy_type: str):
        if energy_type not in self.labels:
            raise ValueError(f"Energy type {energy_type} not available, choose from: {self.labels}")
        i = self.labels.index(energy_type)
        return self.energies[:, i]


class XVGParser(object):

    def __init__(self, path_xvg: str):
        # this is done in order to function with .xvg ending or with no ending
        if not path_xvg.endswith(".xvg"):
            self.path_name = f"{path_xvg}.xvg"
        else:
            self.path_name = path_xvg
        reader = XVGReader(self.path_name)
        self.all_values = reader._auxdata_values
        reader.close()

    def get_parsed_energy(self):
        return ParsedEnergy(self.get_all_columns(), self.get_all_labels(), self.get_y_unit())

    def get_all_columns(self) -> NDArray:
        return self.all_values

    def get_all_labels(self) -> tuple:
        result = []
        with open(self.path_name, "r") as f:
            for line in f:
                # parse column number
                for i in range(0, 10):
                    if line.startswith(f"@ s{i} legend"):
                        split_line = line.split('"')
                        result.append(split_line[-2])
                if not line.startswith("@") and not line.startswith("#"):
                    break
        return tuple(result)

    def get_y_unit(self) -> str:
        y_unit = None
        with open(self.path_name, "r") as f:
            for line in f:
                # parse property unit
                if line.startswith("@    yaxis  label"):
                    split_line = line.split('"')
                    y_unit = split_line[1]
                    break
        if y_unit is None:
            print("Warning: energy units could not be detected in the xvg file.")
            y_unit = "[?]"
        return y_unit

    def get_column_index_by_name(self, column_label) -> Tuple[str, int]:
        correct_column = None
        with open(self.path_name, "r") as f:
            for line in f:
                # parse column number
                if f'"{column_label}"' in line:
                    split_line = line.split(" ")
                    correct_column = int(split_line[1][1:]) + 1
                if not line.startswith("@") and not line.startswith("#"):
                    break
        if correct_column is None:
            print(f"Warning: a column with label {column_label} not found in the XVG file. Using the first y-axis "
                  f"column instead.")
            column_label = "XVG column 1"
            correct_column = 1
        return column_label, correct_column


class ParsedTrajectory:

    def __init__(self, name: str, molecule_generator: Callable, energies: ParsedEnergy, is_pt=False,
                 default_atom_selection=None, dt=1):
        self.name = name
        self.is_pt = is_pt
        self.molecule_generator = molecule_generator
        self.energies = energies
        self.c_num = None
        self.r_num = None
        self.default_atom_selection = default_atom_selection
        self.dt = dt

    def get_num_unique_com(self, energy_type="Potential", atom_selection=None):
        if atom_selection is None:
            atom_selection = self.default_atom_selection
        return len(self.get_unique_com(energy_type=energy_type, atom_selection=atom_selection)[0])

    def assign_coms_2_grid_points(self, full_grid: FullGrid, atom_selection=None, coms=None,
                                  nan_free=False) -> Tuple[NDArray, NDArray]:
        """
        Given a simulation, select centres of mass of a subset of atoms (eg. one of the molecules and assign them
        to the Voronoi cells of a selected position grid. This is useful to see what part of position space a
        classical simulation has covered or to see which molecules moved out of their original grid point area
        after a short optimisation.

        Args:
            full_grid: to which full grid the COMs should be fitted
            atom_selection: used to select the atom belonging to the wished group

        Returns:
            an array of numbers, length the same as the length of the parsed trajectory, each number represents the
            index of the point in the flattened position grid of the full_grid that is closest to this COM.
        """
        if atom_selection is None:
            atom_selection = self.default_atom_selection
        if coms is None:
            coms = self.get_all_COM(atom_selection=atom_selection)
        if nan_free:
            # remove NaNs and convert the rest to int
            return full_grid.nan_free_assignments(coms)
        else:
            return coms, full_grid.point2cell_position_grid(coms)

    def set_c_r(self, c_num: int, r_num: int):
        self.c_num = c_num
        self.r_num = r_num

    def get_all_energies(self, energy_type: str):
        return self.energies.get_energies(energy_type)

    def get_all_COM(self, atom_selection=None) -> NDArray:
        if atom_selection is None:
            atom_selection = self.default_atom_selection
        all_com = []
        for mol in self.molecule_generator(atom_selection):
            all_com.append(mol.get_center_of_mass())
        all_com = np.array(all_com)
        return all_com

    def get_name(self):
        return self.name

    def get_atom_selection_r(self):
        if self.c_num is None or self.r_num is None:
            return None
        return f"id {self.c_num + 1}:{self.c_num + self.r_num}"

    def get_atom_selection_c(self):
        if self.c_num is None or self.r_num is None:
            return None
        return f"id 1:{self.c_num}"

    def get_unique_com(self, energy_type: str = "Potential", atom_selection=None):
        """
        Get only the subset of COMs that have unique positions. Among those, select the ones with lowest energy (if
        energy info is provided)
        """
        if atom_selection is None:
            atom_selection = self.default_atom_selection
        round_to = 3  # number of decimal places
        my_energies = self.energies
        my_coms = self.get_all_COM(atom_selection)

        if my_energies is None or energy_type is None:
            _, indices = np.unique(my_coms.round(round_to), axis=0, return_index=True)
            unique_coms = np.take(my_coms, indices, axis=0)
            return unique_coms, None
        else:
            my_energies = self.energies.get_energies(energy_type)

        # if there are energies, among same COMs, select the one with lowest energy
        coms_tuples = [tuple(row.round(round_to)) for row in my_coms]
        df = pd.DataFrame()
        df["coms_tuples"] = coms_tuples
        df["energy"] = my_energies
        new_df = df.loc[df.groupby(df["coms_tuples"], sort=False)["energy"].idxmin()]
        unique_coms = np.take(my_coms, new_df.index, axis=0)
        unique_energies = np.take(my_energies, new_df.index, axis=0)
        return unique_coms, unique_energies

    def get_only_lowest_highest(self, coms: NDArray, energies: NDArray, lowest_k: int = None, highest_j: int = None):
        """
        After any step that produces a set of COMs and corresponding energies, you may filter out only the (COM, energy)
        with k lowest and/or j highest energies. This is useful for plotting so that only relevant points are visible.
        """
        if lowest_k or highest_j:
            # if not set, just set it to zero
            if lowest_k is None:
                lowest_k = 0

            order = energies.argsort()
            coms = coms[order]
            energies = energies[order]
            # use lowest values
            selected_coms = coms[:lowest_k]
            selected_energies = energies[:lowest_k]
            if highest_j is not None:
                selected_coms = np.concatenate((selected_coms, coms[-highest_j:]))
                selected_energies = np.concatenate((selected_energies, energies[-highest_j:]))
            return selected_coms, selected_energies
        return coms, energies

    def get_unique_com_till_N(self, N: int, energy_type: str = "Potential", atom_selection=None):
        if atom_selection is None:
            atom_selection = self.default_atom_selection
        coms, ens = self.get_unique_com(energy_type=energy_type, atom_selection=atom_selection)
        if ens is None:
            return coms[:N], None
        return coms[:N], ens[:N]


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    # SIMULATION
    path_energy = "/home/mdglasius/Modelling/trypsin_normal/nobackup/outputs/data.txt"
    df = pd.read_csv(path_energy)
    energies = df['Potential Energy (kJ/mole)'].to_numpy()[:, np.newaxis]
    pe = ParsedEnergy(energies=energies, labels=["Potential Energy"], unit="(kJ/mole)")
    pt_parser = FileParser(
             path_topology="/home/mdglasius/Modelling/trypsin_normal/inputs/trypsin_probe.pdb",
             path_trajectory="/home/mdglasius/Modelling/trypsin_normal/nobackup/outputs/aligned_traj.dcd")
    print(pt_parser.get_dt())
    parsed_trajectory = pt_parser.get_parsed_trajectory()
    parsed_trajectory.energies = pe
    #fg = FullGrid(t_grid_name="linspace(0.8, 2.5, 8)", o_grid_name="ico_512", b_grid_name="zero")
    fg = FullGrid(t_grid_name="[5, 10, 15]", o_grid_name="ico_100", b_grid_name="zero")
    coms, assignments = parsed_trajectory.assign_coms_2_grid_points(full_grid=fg, atom_selection="segid B")
    num_cells_visited = len(np.unique(assignments))
    num_cells_total = len(fg.get_flat_position_grid())
    print(f"Percentage of non-empty cells: {np.round(num_cells_visited/num_cells_total * 100, 2)}%")
    # import seaborn as sns
    # import matplotlib.pyplot as plt
    # bins = np.arange(0, len(fg.get_flat_position_grid()))
    # sns.histplot(assignments, bins=bins)
    # sns.scatterplot(x=assignments, y=energies.squeeze(), ax=plt.gca().twinx())
    # plt.show()
    # from molgri.plotting.molecule_plots import TrajectoryPlot
    # from molgri.plotting.fullgrid_plots import FullGridPlot
    # tp = TrajectoryPlot(parsed_trajectory)
    # #tp.make_COM_plot(atom_selection="segid B", animate_rot=True)
    # tp.make_energy_COM_plot(atom_selection="segid B", animate_rot=False, energy_type="Potential Energy",
    #                         projection="3d", save=False)
    # fg = FullGridPlot(fg)
    # fg.make_full_voronoi_plot(ax=tp.ax, fig=tp.fig, animate_rot=True, plot_vertex_points=False)


"""
The sub-package molgri.molecules is largely based on MDAnalysis handling of molecular structure and trajectory files.
However, the parsers are wrapped in own objects and passed to Pseudotrajectory generators that write out a sequence
of structures according to FullGrid prescription. Analysis of (pseudo)trajectories is possible with the module
transitions.
"""


"""
Plots connected to the fullgrid module.

Plot position grids in space, Voronoi cells and their volumes etc.
"""

import numpy as np
import seaborn as sns
from matplotlib import colors

from molgri.constants import GRID_ALGORITHMS, NAME2SHORT_NAME
from molgri.plotting.abstract import RepresentationCollection, PanelRepresentationCollection, plot_voronoi_cells
from molgri.space.fullgrid import FullGrid, ConvergenceFullGridO


class FullGridPlot(RepresentationCollection):

    def __init__(self, full_grid: FullGrid, *args, **kwargs):
        self.full_grid = full_grid
        self.full_voronoi_grid = full_grid.get_full_voronoi_grid()
        data_name = self.full_grid.get_name()
        super().__init__(data_name, *args, **kwargs)

    def get_possible_title(self, algs = True, N_points = False):
        name = ""
        if algs:
            o_name = NAME2SHORT_NAME[self.full_grid.o_rotations.algorithm_name]
            b_name = NAME2SHORT_NAME[self.full_grid.b_rotations.algorithm_name]
            name += f"o = {o_name}, b = {b_name}"
        if N_points:
            N_o = self.full_grid.o_rotations.N
            N_b = self.full_grid.b_rotations.N
            N_t = self.full_grid.t_grid.get_N_trans()
            N_name = f"N_o = {N_o}, N_b = {N_b}, N_t = {N_t}"
            if len(name) > 0:
                N_name = f"; {N_name}"
            name += N_name
        return name

    def make_position_plot(self, ax=None, fig=None, save=True, animate_rot=False, numbered: bool = False,
                           c="black", projection="3d"):
        self._create_fig_ax(fig=fig, ax=ax, projection=projection)

        points = self.full_grid.get_flat_position_grid()
        cmap = "bwr"
        norm = colors.TwoSlopeNorm(vcenter=0)
        self.ax.scatter(*points.T, c=c, cmap=cmap, norm=norm)

        if numbered:
            for i, point in enumerate(points):
                self.ax.text(*point, s=f"{i}")

        if projection == "3d":
            self.ax.view_init(elev=10, azim=30)
            self._set_axis_limits()
            self._equalize_axes()

        if save:
            self._save_plot_type(f"position_{projection}")
        if animate_rot:
            return self._animate_figure_view(self.fig, self.ax, f"position_rotated")

    def make_full_voronoi_plot(self, ax=None, fig=None, save=True, animate_rot=False, plot_vertex_points=True,
                               numbered: bool = False):
        self._create_fig_ax(fig=fig, ax=ax, projection="3d")

        origin = np.zeros((3,))

        if numbered:
            points = self.full_grid.get_flat_position_grid()
            for i, point in enumerate(points):
                self.ax.text(*point, s=f"{i}")

        try:
            voronoi_disc = self.full_voronoi_grid.get_voronoi_discretisation()

            for i, sv in enumerate(voronoi_disc):
                plot_voronoi_cells(sv, self.ax, plot_vertex_points=plot_vertex_points)
                # plot rays from origin to highest level
                if i == len(voronoi_disc)-1:
                    for vertex in sv.vertices:
                        ray_line = np.concatenate((origin[:, np.newaxis], vertex[:, np.newaxis]), axis=1)
                        self.ax.plot(*ray_line, color="black")
        except AttributeError:
            pass

        self.ax.view_init(elev=10, azim=30)
        self._set_axis_limits()
        self._equalize_axes()

        if save:
            self._save_plot_type("full_voronoi")
        if animate_rot:
            return self._animate_figure_view(self.fig, self.ax, f"full_voronoi_rotated")

    def make_point_vertices_plot(self, point_index: int, ax=None, fig=None, save=True, animate_rot=False,
                                 which="all"):
        self.make_full_voronoi_plot(ax=ax, fig=fig, save=False, plot_vertex_points=False)
        self.make_position_plot(save=False, numbered=True, ax=self.ax, fig=self.fig)

        try:
            vertices = self.full_voronoi_grid.find_voronoi_vertices_of_point(point_index, which=which)
            self.ax.scatter(*vertices.T, color="red")
        except AttributeError:
            pass

        save_name = f"vertices_of_{point_index}"
        if save:
            self._save_plot_type(save_name)
        if animate_rot:
            return self._animate_figure_view(self.fig, self.ax, save_name)

    def create_all_plots(self, and_animations=False):
        self.make_position_plot(numbered=True, animate_rot=and_animations)
        self.make_position_plot(numbered=False, animate_rot=and_animations)
        self.make_full_voronoi_plot(plot_vertex_points=True, animate_rot=and_animations)
        self.make_full_voronoi_plot(plot_vertex_points=False, animate_rot=and_animations)
        self.make_point_vertices_plot(0, animate_rot=and_animations)


class ConvergenceFullGridPlot(RepresentationCollection):

    def __init__(self, convergence_full_grid: ConvergenceFullGridO):
        self.convergence_full_grid = convergence_full_grid
        super().__init__(self.convergence_full_grid.get_name())

    def get_possible_title(self):
        full_name = self.convergence_full_grid.get_name()
        split_name = full_name.split("_")
        return NAME2SHORT_NAME[split_name[2]]

    def make_voronoi_volume_conv_plot(self, ax=None, fig=None, save=True):

        self._create_fig_ax(fig=fig, ax=ax)

        try:
            voronoi_df = self.convergence_full_grid.get_voronoi_volumes()

            all_layers = set(voronoi_df["layer"])

            for layer in all_layers:
                filtered_df = voronoi_df.loc[voronoi_df['layer'] == layer]
                sns.lineplot(data=filtered_df, x="N", y="Voronoi cell volume", errorbar="sd", ax=self.ax)
                sns.scatterplot(data=filtered_df, x="N", y="Voronoi cell volume", alpha=0.8, color="black", ax=self.ax,
                                s=1)
                sns.scatterplot(data=filtered_df, x="N", y="ideal volume", color="black", marker="x", ax=self.ax)
        except AttributeError:
            pass

        if save:
            self.ax.set_xscale("log")
            self._save_plot_type("voronoi_volume_conv")


class PanelConvergenceFullGridPlots(PanelRepresentationCollection):

    def __init__(self, t_grid_name: str, b_grid_name: str = "zero", N_set: list = None, **kwargs):
        list_plots = []
        for alg in GRID_ALGORITHMS[:-1]:
            conv_sphere_grid = ConvergenceFullGridO(o_alg_name=alg, N_set=N_set, b_grid_name=b_grid_name,
                                                    t_grid_name=t_grid_name, filter_non_unique=True, **kwargs)
            sphere_plot = ConvergenceFullGridPlot(conv_sphere_grid)
            list_plots.append(sphere_plot)
        data_name = f"all_convergence_{list_plots[0].convergence_full_grid.get_name()}"
        super().__init__(data_name, list_plots)

    def make_all_voronoi_volume_plots(self, save=True):
        self._make_plot_for_all("make_voronoi_volume_conv_plot")
        self.add_titles(list_titles=[subplot.get_possible_title() for subplot in self.list_plots])
        self.set_log_scale(x_axis=True, y_axis=False)
        self.unify_axis_limits()
        if save:
            self._save_multiplot("voronoi_volume")


if __name__ == "__main__":
    from molgri.constants import SMALL_NS, DEFAULT_NS, MINI_NS
    fg = FullGrid("zero", "ico_42", "[0.1, 0.2, 0.3, 0.4, 0.5]") #
    fgp = FullGridPlot(fg, default_complexity_level="half_empty", default_context="talk")
    fgp.make_position_plot()
    # fgp.make_full_voronoi_plot(ax=fgp.ax, fig=fgp.fig, animate_rot=True, numbered=True, plot_vertex_points=False)
    #cfgo = ConvergenceFullGridO("zero", t_grid_name="[0.1, 0.3]", o_alg_name="ico", use_saved=True)
    #ConvergenceFullGridPlot(cfgo).make_voronoi_volume_conv_plot()
    # PanelConvergenceFullGridPlots(t_grid_name="[1.5, 3]", use_saved=False,
    #                               N_set=SMALL_NS).make_all_voronoi_volume_plots()
    # PanelConvergenceFullGridPlots(t_grid_name="[1.5, 3]", N_set=DEFAULT_NS, use_saved=False).make_all_voronoi_volume_plots()


"""
Logging. Naming of files with first free index.
"""

import os
from abc import ABC, abstractmethod
import logging


class AbstractLogger(ABC):

    def __init__(self, path: str, level: str = "INFO"):
        class_name = type(self).__name__
        logging.basicConfig(filename=path, level=level)
        self.logger = logging.getLogger(class_name)

    @abstractmethod
    def log_set_up(self, investigated_object: object):
        self.logger.info(f"SET UP OF: {investigated_object}")


class PtLogger(AbstractLogger):

    def log_set_up(self, investigated_object, ):
        super(PtLogger, self).log_set_up(investigated_object)


def first_index_free_4_all(list_names, list_endings, list_paths, index_places: int = 4) -> int:
    """Similar to find_first_free_index, but you have a list of files (eg. trajectory, topology, log) -> all must
    be still free to use this specific index."""
    assert len(list_paths) == len(list_names) == len(list_endings)
    i = 0
    # the loop only makes sense till you use up all numbers that could be part of the name
    while i < 10**index_places:
        for name, ending, path in zip(list_names, list_endings, list_paths):
            # if any one of them exist, break
            if os.path.exists(format_name(file_path=path, file_name=name, num=i, places_num=index_places,
                                          suffix=ending)):
                i += 1
                break
        # if you did not break the loop, all of the indices are free
        else:
            return i
    raise FileNotFoundError(f"All file names with unique numbers {format(0, f'0{index_places}d')}-{10**index_places-1} "
                            f"are already used up!")


def paths_free_4_all(list_names, list_endings, list_paths, index_places: int = 4) -> tuple:
    num = first_index_free_4_all(list_names, list_endings, list_paths, index_places)
    result_paths = []
    for name, ending, path in zip(list_names, list_endings, list_paths):
        result = format_name(file_name=name, file_path=path, num=num, places_num=index_places, suffix=ending)
        result_paths.append(result)
    return tuple(result_paths)


def find_first_free_index(name: str, ending: str = None, index_places: int = 4, path: str = "") -> int:
    """
    Problem: you want to save a file with a certain name, eg. pics/PrettyFigure.png, but not overwrite any existing
    files.

    Solution: this function checks if the given file already exists and increases the index until a unique one is
    found. So for our example it may return the string pics/PrettyFigure_007.png

    If no such file exists yet, the return string will include as many zeros as defined by index_places

    Args:
        path: path to appropriate directory (if not current), ending with /, in the example 'pics/'
        name: name of the file, in the example 'PrettyFigure'
        ending: in the example, select 'png' (NO DOT)
        index_places: how many places are used for the number

    Returns:
        number - first free index (can be forwarded to format_name)
    """
    i = 0
    while os.path.exists(format_name(file_path=path, file_name=name, num=i, places_num=index_places, suffix=ending)):
        i += 1
    return i


def format_name(file_name: str, num: int = None, places_num: int = 4, suffix: str = None, file_path: str = ""):
    """

    Args:
        file_path: eg. output/
        file_name: eg my_file
        num: eg 17
        places_num: eg 4 -> num will be formatted as 0017
        suffix: ending of the file

    Returns:
        full path and file name in correct format
    """
    till_num = f"{file_path}{file_name}"
    if num is None:
        till_ending = till_num
    else:
        till_ending = f"{till_num}_{format(num, f'0{places_num}d')}"
    if suffix is None:
        return till_ending
    return f"{till_ending}.{suffix}"


"""
Parse linear discretisation in units provided by users.

Able to process strings like '[1, 2, 3]', 'linspace(1, 5, 50)' or 'range(0.5, 3, 0.4)' and translate them to appropriate
arrays of distances.
"""

import hashlib
import numbers
from ast import literal_eval

import numpy as np
from numpy.typing import NDArray

from molgri.constants import NM2ANGSTROM
from molgri.paths import PATH_OUTPUT_TRANSGRIDS


class TranslationParser(object):

    """
    User input is expected in nanometers (nm)!

        Parse all ways in which the user may provide a linear translation grid. Currently supported formats:
            - a list of numbers, eg '[1, 2, 3]'
            - a linearly spaced list with optionally provided number of elements eg. 'linspace(1, 5, 50)'
            - a range with optionally provided step, eg 'range(0.5, 3, 0.4)'
    """

    def __init__(self, user_input: str):
        """
        Args:
            user_input: a string in one of allowed formats
        """
        self.user_input = user_input
        if "linspace" in self.user_input:
            bracket_input = self._read_within_brackets()
            self.trans_grid = np.linspace(*bracket_input, dtype=float)
        elif "range" in self.user_input:
            bracket_input = self._read_within_brackets()
            self.trans_grid = np.arange(*bracket_input, dtype=float)
        else:
            self.trans_grid = literal_eval(self.user_input)
            self.trans_grid = np.array(self.trans_grid, dtype=float)
            self.trans_grid = np.sort(self.trans_grid, axis=None)
        # convert to angstrom
        self.trans_grid = self.trans_grid * NM2ANGSTROM
        # we use a (shortened) hash value to uniquely identify the grid used, no matter how it's generated
        self.grid_hash = int(hashlib.md5(self.trans_grid).hexdigest()[:8], 16)
        # save the grid (for record purposes)
        path = f"{PATH_OUTPUT_TRANSGRIDS}trans_{self.grid_hash}.txt"
        # noinspection PyTypeChecker
        np.savetxt(path, self.trans_grid)

    def get_trans_grid(self) -> NDArray:
        """Getter to access all distances from origin in angstorms."""
        return self.trans_grid

    def get_N_trans(self) -> int:
        """Get the number of translations in this grid."""
        return len(self.trans_grid)

    def sum_increments_from_first_radius(self) -> float:
        """
        Get final distance - first non-zero distance == sum(increments except the first one).

        Useful because often the first radius is large and then only small increments are made.
        """
        return float(np.sum(self.get_increments()[1:]))

    def get_increments(self) -> NDArray:
        """
        Get an array where each element represents an increment needed to get to the next radius.

        Example:
            self.trans_grid = np.array([10, 10.5, 11.2])
            self.get_increments() -> np.array([10, 0.5, 0.7])
        """
        increment_grid = [self.trans_grid[0]]
        for start, stop in zip(self.trans_grid, self.trans_grid[1:]):
            increment_grid.append(stop-start)
        increment_grid = np.array(increment_grid)
        assert np.all(increment_grid > 0), "Negative or zero increments in translation grid make no sense!"
        return increment_grid

    def _read_within_brackets(self) -> tuple:
        """
        Helper function to aid reading linspace(start, stop, num) and arange(start, stop, step) formats.
        """
        str_in_brackets = self.user_input.split('(', 1)[1].split(')')[0]
        str_in_brackets = literal_eval(str_in_brackets)
        if isinstance(str_in_brackets, numbers.Number):
            str_in_brackets = tuple((str_in_brackets,))
        return str_in_brackets



"""
The sub-package molgri.space is only concerned with geometry in 3- and 4-dimensional spaces. It includes general
functionality and no references to molecular structures.

Most of the modules are intended for internal use. The module that is the interface between grid implementation and
pseudo-trajectory implementation is the fullgrid module.

Modules:
 - analysis: implements statistical tests of uniformity in 3D and 4D spherical grids
 - **fullgrid: combines position grid with orientation grid and thus fully discretises approach space in Voronoi cells**
 - polytopes: implements 3D and 4D polytopes: cube, hypercube and icosahedron - needed for rotobj module
 - rotations: implements transformation between grid objects and rotation objects
 - **rotobj: implements SphereGridNDim object using varius algorithms to distribute points on N-dim sphere**
 - translations: parses user inputs of translations (linear space discretisation)
 - utils: provides other useful functions, eg normalisation, random quaternions
"""


"""
User script for generating custom rotational grids.
"""

import argparse

from molgri.space.rotobj import SphereGridFactory
from molgri.plotting.spheregrid_plots import SphereGridPlot
from molgri.scripts.set_up_io import freshly_create_all_folders

import warnings
warnings.filterwarnings("ignore")

parser = argparse.ArgumentParser()
requiredNamed = parser.add_argument_group('required named arguments')
requiredNamed.add_argument('-N', metavar='N', type=int, nargs='?', required=True,
                           help='the number of points in your rotational grid')
requiredNamed.add_argument('-algorithm', metavar='a', type=str, nargs='?', required=True,
                           help='name of the grid-generating algorithm to use'
                                ' (ico, cube3D, cube4D, randomQ, randomE, systemE)')
parser.add_argument('--recalculate', action='store_true',
                    help='recalculate the grid even if a saved version already exists')
parser.add_argument('--dimensions', type=int, default=3,
                    help='select 3 for grids on a sphere in 3D space and 4 for grids on a hypersphere in 4D space')
parser.add_argument('--statistics', action='store_true',
                    help='write out statistics and draw uniformity and convergence plots')
parser.add_argument('--draw', action='store_true',
                    help='draw the grid and save the plot')
parser.add_argument('--background', action='store_true', default=True,
                    help='when drawing a grid, display axes and ticks.')
parser.add_argument('--animate', action='store_true',
                    help='provide an animation of the grid rotating in 3D')
parser.add_argument('--animate_ordering', action='store_true',
                    help='provide an animation of the grid generation')
parser.add_argument('--animate_translation', action='store_true',
                    help='provide an animation of the grid sliding through one of the dimensions')
parser.add_argument('--readable', action='store_true',
                    help='save the grid file in a txt format as well')


def prepare_grid(args):
    algo = args.algorithm
    n_points = args.N
    # if already exists and no --recalculate flag, just display a message
    use_saved = not args.recalculate
    my_sphere_grid = SphereGridFactory.create(N=n_points, alg_name=algo, dimensions=args.dimensions,
                                              use_saved=use_saved, time_generation=True, print_messages=True)
    my_sphere_grid.save_grid()
    print(f"The grid can be found at {my_sphere_grid.get_grid_path()}")
    # if running from another script, args may not include the readable attribute
    try:
        if args.readable:
            extension = "txt"
            my_sphere_grid.save_grid(extension=extension)
            path = my_sphere_grid.get_grid_path(extension=extension)
            print(f"A .txt version of grid can be found at {path}.")
    except AttributeError:
        pass
    return my_sphere_grid


def run_generate_grid():
    freshly_create_all_folders()
    my_args = parser.parse_args()
    grid_name = f"{my_args.algorithm}_{my_args.N}"
    my_rotations = prepare_grid(my_args)

    # any plotting is done with this object
    default_complexity_level = "full"
    if not my_args.background:
        default_complexity_level = "empty"
    sgp = SphereGridPlot(my_rotations, default_context="talk", default_complexity_level=default_complexity_level)

    if my_args.animate or my_args.animate_ordering:
        my_args.draw = True
    if my_args.draw:
        sgp.make_grid_plot()
        print(f"Grid drawn and figure saved to {sgp.fig_path}.")
        if my_args.animate:
            sgp.make_rot_animation()
            print(f"Animation of the grid saved to {sgp.ani_path}")
        if my_args.animate_ordering:
            sgp.make_ordering_animation()
            print(f"Animation of the grid ordering saved to {sgp.ani_path}")
        if my_args.animate_translation:
            sgp.make_trans_animation()
            print(f"Animation of the grid sliding through the last dimension saved to {sgp.ani_path}")
    if my_args.statistics:
        my_rotations.save_uniformity_statistics()
        print(f"A statistics file describing the grid {grid_name} was saved to {my_rotations.get_statistics_path('csv')}.")
        sgp.make_uniformity_plot()
        print(f"A violin plot showing the uniformity of {grid_name} saved to {sgp.fig_path}.")
        sgp.make_convergence_plot()
        print(f"A convergence plot with number of points between 3 and {my_args.N} saved to {sgp.fig_path}.")


if __name__ == '__main__':
    run_generate_grid()
    print("Generation finished.")



"""
Grids on N-dimensional spheres.

This module is one of central building blocks of the molgri package. The central object here is the SphereGridNDim
(And the corresponding Factory object that creates a sphere grid given the algorithm, dimension and number of points).
A SphereGridNDim object implements a grid that consists of points on a N-dimensional unit sphere.

Connected to:
 - polytope module - some algorithms for creating sphere grids are based of subdivision of polytopes
 - fullgrid module - combines a 3D SphereGridNDim, a 4D SphereGridNDim and a translation grid to sample the full
                     approach space
"""

import os
from abc import ABC, abstractmethod
from time import time
from typing import Optional

import numpy as np
import pandas as pd
from numpy.typing import NDArray
from scipy.constants import pi
from scipy.spatial import SphericalVoronoi
from scipy.spatial.transform import Rotation

from molgri.space.analysis import prepare_statistics, write_statistics
from molgri.space.utils import random_quaternions, standardise_quaternion_set
from molgri.constants import UNIQUE_TOL, EXTENSION_GRID_FILES, NAME2PRETTY_NAME, SMALL_NS
from molgri.paths import PATH_OUTPUT_ROTGRIDS, PATH_OUTPUT_STAT
from molgri.space.polytopes import Cube4DPolytope, IcosahedronPolytope, Cube3DPolytope, Polytope
from molgri.space.rotations import grid2rotation, rotation2grid4vector
from molgri.wrappers import time_method, save_or_use_saved, deprecated


class SphereGridNDim(ABC):
    """
    This is a general and abstract implementation of a spherical grid in any number of dimensions (currently only
    using 3 or 4 dimensions). Each subclass is a particular implementation that must implement the abstract method
    _gen_grid_4D. If the implementation of _gen_grid_3D is not overridden, the 3D grid will be created from the 4D
    one using 4D grid to act as rotational quaternions on the unit z vector.
    """

    algorithm_name = "generic"

    def __init__(self, dimensions: int, N: int = None, use_saved: bool = True,
                 print_messages: bool = False, time_generation: bool = False, filter_non_unique: bool = False):
        self.dimensions = dimensions
        self.N = N
        self.gen_algorithm = self.algorithm_name
        self.use_saved = use_saved
        self.time_generation = time_generation
        self.print_messages = print_messages
        self.grid: Optional[NDArray] = None
        self.filter_non_unique = filter_non_unique
        self.spherical_voronoi: Optional[SphericalVoronoi] = None

    def __len__(self) -> int:
        return self.get_N()

    def get_N(self) -> int:
        """Get the number of points in the self.grid array. It is important to use this getter and not the attribute
        self.N as they can differ if the self.filter_non_unique option is set."""
        return len(self.get_grid_as_array())

    def __str__(self):
        return f"Object {type(self).__name__} <{self.get_decorator_name()}>"

    ##################################################################################################################
    #                      generation/loading of grids
    ##################################################################################################################

    def gen_grid(self) -> NDArray:
        """
        This method saves to self.grid if it has been None before (and checks the format) but returns nothing.
        This method only implements loading/timing/printing logic, the actual process of creation is outsourced to
        self._gen_grid() that is implemented by sub-classes.

        To get the grid, use self.get_grid_as_array().
        """
        # condition that there is still something to generate
        if self.grid is None:
            if self.time_generation:
                self.grid = self.gen_and_time()
            else:
                self.grid = self._gen_grid()
            if self.print_messages:
                self._check_uniqueness()
        assert isinstance(self.grid, np.ndarray), "A grid must be a numpy array!"
        if not self.filter_non_unique:
            assert self.grid.shape == (self.N, self.dimensions), f"Grid not of correct shape!"
        assert np.allclose(np.linalg.norm(self.grid, axis=1), 1, atol=10 ** (-UNIQUE_TOL)), "A grid must have norm 1!"
        return self.grid

    def _gen_grid(self) -> NDArray:
        if self.dimensions == 4:
            return self._gen_grid_4D()
        elif self.dimensions == 3:
            return self._gen_grid_3D()
        else:
            raise NotImplementedError(f"Generating sphere grids in {self.dimensions} not implemented!")

    def _gen_grid_3D(self) -> NDArray:
        """
        This is the default of obtaining 3D grid by projecting on z-vector, should be overwritten for ico and cube3D
        """
        quaternions = self._gen_grid_4D()
        rotations = Rotation.from_quat(quaternions)
        points = rotation2grid4vector(rotations)
        if self.filter_non_unique:
            points = provide_unique(points)
        return points

    @abstractmethod
    def _gen_grid_4D(self) -> NDArray:
        pass

    @time_method
    def gen_and_time(self) -> NDArray:
        """Same as generation, but prints the information about time needed. Useful for end users."""
        return self._gen_grid()

    def _check_uniqueness(self):
        """
        Optional - check if every grid point in this grid is unique (not too close to other points). Will fail for
        sufficiently dense grids and same algorithms.
        """
        as_quat = False
        if self.dimensions == 4:
            as_quat = True
        unique_grid = provide_unique(self.grid, as_quat=as_quat)
        if len(self.grid) != len(unique_grid):
            print(f"Warning! {len(self.grid) - len(unique_grid)} grid points of "
                  f"{self.get_name(with_dim=True)} non-unique (distance < 10^-{UNIQUE_TOL}).")

    ##################################################################################################################
    #                      name and path getters
    ##################################################################################################################

    def get_name(self, with_dim: bool = True) -> str:
        """
        This is a standard name that can be used for saving files, but not pretty enough for titles etc.

        Args:
            with_dim: if True, include _3d or _4d to include dimensionality of the grid

        Returns:
            a string usually of the form 'ico_17'
        """
        output = f"{self.gen_algorithm}"
        if self.N is not None:
            output += f"_{self.N}"
        if with_dim:
            output += f"_{self.dimensions}d"
        return output

    def get_decorator_name(self) -> str:
        """Name used in printing and decorators, not suitable for file names."""
        return f"{NAME2PRETTY_NAME[self.gen_algorithm]} algorithm, {self.N} points"

    def get_grid_path(self, extension=EXTENSION_GRID_FILES) -> str:
        """Get the entire path to where the fullgrid is saved."""
        return f"{PATH_OUTPUT_ROTGRIDS}{self.get_name(with_dim=True)}.{extension}"

    def get_statistics_path(self, extension) -> str:
        """get the entire path to where the statistics are saved."""
        return f"{PATH_OUTPUT_STAT}{self.get_name(with_dim=True)}.{extension}"

    ##################################################################################################################
    #                      useful methods
    ##################################################################################################################

    @save_or_use_saved
    def get_grid_as_array(self) -> NDArray:
        """
        Get the sphere grid. If not yet created but N is set, will generate/load the grid automatically.
        """
        self.gen_grid()
        return self.grid

    @deprecated
    def save_grid(self, extension: str = EXTENSION_GRID_FILES):
        """
        Deprecated - all necessary data is saved by individual methods using save_or_use_saved.
        """
        if extension == "txt":
            # noinspection PyTypeChecker
            np.savetxt(self.get_grid_path(extension=extension), self.get_grid_as_array())
        else:
            np.save(self.get_grid_path(extension=extension), self.get_grid_as_array())

    @deprecated
    def save_uniformity_statistics(self, num_random: int = 100, alphas=None):
        """
        Deprecated - all necessary data is saved by individual methods using save_or_use_saved.
        Save both the short, user-friendly summary of statistics and the entire uniformity data.
        """
        short_statistics_path = self.get_statistics_path(extension="txt")
        statistics_path = self.get_statistics_path(extension="csv")
        stat_data, full_data = prepare_statistics(self.get_grid_as_array(), alphas, d=self.dimensions,
                                                  num_rand_points=num_random)
        write_statistics(stat_data, full_data, short_statistics_path, statistics_path,
                         num_random, name=self.get_name(), dimensions=self.dimensions,
                         print_message=self.print_messages)

    @save_or_use_saved
    def get_uniformity_df(self, alphas):
        """
        Get the dataframe necessary to draw violin plots showing how uniform different generation algorithms are.
        """
        # recalculate if: 1) self.use_saved_data = False OR 2) no saved data exists
        if not self.use_saved or not os.path.exists(self.get_statistics_path("csv")):
            self.save_uniformity_statistics(alphas=alphas)
        ratios_df = pd.read_csv(self.get_statistics_path("csv"), dtype=float)
        # OR 3) provided alphas don't match those in the found file
        saved_alphas = set(ratios_df["alphas"])
        if saved_alphas != alphas:
            self.save_uniformity_statistics(alphas=alphas)
            ratios_df = pd.read_csv(self.get_statistics_path("csv"), dtype=float)
        return ratios_df

    @save_or_use_saved
    def get_convergence_df(self, alphas: tuple, N_list: tuple = None):
        """
         Get the dataframe necessary to draw convergence plots for various values of N.
         """
        if N_list is None:
            # create equally spaced convergence set
            assert self.N >= 3, f"N={self.N} not large enough to study convergence"
            N_list = np.logspace(np.log10(3), np.log10(self.N), dtype=int)
            N_list = np.unique(N_list)
        full_df = []
        for N in N_list:
            grid_factory = SphereGridFactory.create(alg_name=self.gen_algorithm, N=N, dimensions=self.dimensions,
                                                    print_messages=False, time_generation=False,
                                                    use_saved=self.use_saved)
            df = grid_factory.get_uniformity_df(alphas=alphas)
            df["N"] = len(grid_factory.get_grid_as_array())
            full_df.append(df)
        full_df = pd.concat(full_df, axis=0, ignore_index=True)
        return full_df

    @save_or_use_saved
    def get_spherical_voronoi_cells(self):
        """
        A spherical grid (in 3D) can be used as a basis for a spherical Voronoi grid. In this case, each grid point is
        used as a center of a Voronoi cell. The spherical Voronoi cells also have the radius 1.

        The division into cells will fail in case points are not unique (too close to each other)
        """
        assert self.dimensions == 3, "Spherical voronoi cells only available for N=3"
        if self.spherical_voronoi is None:
            try:
                self.spherical_voronoi = SphericalVoronoi(self.grid, radius=1, threshold=10**-UNIQUE_TOL)
            except ValueError:
                if self.print_messages:
                    print(f"Spherical voronoi not created for {self.get_name()} due to duplicate generators")
                self.spherical_voronoi = []
        return self.spherical_voronoi

    @save_or_use_saved
    def get_voronoi_areas(self):
        """
        From Voronoi cells you may also calculate areas on the sphere that are closest each grid point. The order of
        areas is the same as the order of points in self.grid.
        """
        sv = self.get_spherical_voronoi_cells()
        return sv.calculate_areas()


def _select_unique_rotations(rotations):
    rot_matrices = rotations.as_matrix()
    rot_matrices = np.unique(np.round(rot_matrices, UNIQUE_TOL), axis=0)
    return Rotation.from_matrix(rot_matrices)


def provide_unique(el_array: NDArray, tol: int = UNIQUE_TOL, as_quat=False) -> NDArray:
    """
    Take an array of any shape in which each row is an element. Return only the elements that are unique up to
    the tolerance.
    # TODO: maybe you could do some sort of decomposition to determine the point of uniqueness?

    Args:
        el_array: array with M elements of any shape
        tol: number of decimal points to consider
        as_quat: select True if el_array is an array of quatenions and you want to consider q and -q to be non-unique
                 (respecting double-coverage of quaternions)

    Returns:
        array with N <= M elements of el_array that are all unique
    """
    if as_quat:
        assert el_array.shape[1] == 4
        _, indices = np.unique(standardise_quaternion_set(el_array.round(tol)), axis=0, return_index=True)
    else:
        _, indices = np.unique(el_array.round(tol), axis=0, return_index=True)
    return np.array([el_array[index] for index in sorted(indices)])


class RandomQRotations(SphereGridNDim):

    algorithm_name = "randomQ"

    def _gen_grid_4D(self) -> NDArray:
        return random_quaternions(self.N)


class SystemERotations(SphereGridNDim):

    algorithm_name = "systemE"

    def _gen_grid_4D(self) -> NDArray:
        num_points = 1
        rot_matrices = []
        while len(rot_matrices) < self.N:
            phis = np.linspace(0, 2 * pi, num_points)
            thetas = np.linspace(0, 2 * pi, num_points)
            psis = np.linspace(0, 2 * pi, num_points)
            euler_meshgrid = np.array(np.meshgrid(*(phis, thetas, psis)), dtype=float)
            euler_meshgrid = euler_meshgrid.reshape((3, -1)).T
            rotations = Rotation.from_euler("ZYX", euler_meshgrid)
            # remove non-unique rotational matrices
            rotations = _select_unique_rotations(rotations)
            rot_matrices = rotations.as_matrix()
            num_points += 1
        # maybe just shuffle rather than order
        np.random.shuffle(rot_matrices)
        # convert to a grid
        rotations = Rotation.from_matrix(rot_matrices[:self.N])
        return rotations.as_quat()


class RandomERotations(SphereGridNDim):

    algorithm_name = "randomE"

    def _gen_grid_4D(self) -> NDArray:
        euler_angles = 2 * pi * np.random.random((self.N, 3))
        rotations = Rotation.from_euler("ZYX", euler_angles)
        return rotations.as_quat()


class ZeroRotations(SphereGridNDim):

    algorithm_name = "zero"

    def _gen_grid_4D(self):
        self.N = 1
        rot_matrix = np.eye(3)
        rot_matrix = rot_matrix[np.newaxis, :]
        self.rotations = Rotation.from_matrix(rot_matrix)
        return self.rotations.as_quat()


class Cube4DRotations(SphereGridNDim):

    algorithm_name = "cube4D"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.polytope = Cube4DPolytope()

    def _gen_grid_4D(self):
        while len(self.polytope.get_node_coordinates()) < self.N:
            self.polytope.divide_edges()
        return self.polytope.get_N_ordered_points(self.N)


class IcoAndCube3DRotations(SphereGridNDim):

    algorithm_name: str = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.polytope: Polytope = None

    def get_z_projection_grid(self):
        "In this case, directly construct the 3D object"

    def _gen_grid_3D(self):
        while len(self.polytope.get_node_coordinates()) < self.N:
            self.polytope.divide_edges()
        ordered_points = self.polytope.get_N_ordered_points(self.N)
        return ordered_points

    def _gen_grid_4D(self):
        grid_z_arr = self._gen_grid_3D()
        rotations = grid2rotation(grid_z_arr, grid_z_arr, grid_z_arr)
        return rotations.as_quat()


class IcoRotations(IcoAndCube3DRotations):

    algorithm_name = "ico"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.polytope = IcosahedronPolytope()


class Cube3DRotations(IcoAndCube3DRotations):

    algorithm_name = "cube3D"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.polytope = Cube3DPolytope()


class SphereGridFactory:

    """
    This should be the only access point to all SphereGridNDim objects. Simply provide the generation algorithm name,
    the number of points and dimensions as well as optional arguments. The requested object will be returned.
    """

    @classmethod
    def create(cls, alg_name: str, N: int, dimensions: int, **kwargs) -> SphereGridNDim:
        if alg_name == "randomQ":
            selected_sub_obj = RandomQRotations(N=N, dimensions=dimensions, **kwargs)
        elif alg_name == "systemE":
            selected_sub_obj = SystemERotations(N=N, dimensions=dimensions, **kwargs)
        elif alg_name == "randomE":
            selected_sub_obj = RandomERotations(N=N, dimensions=dimensions, **kwargs)
        elif alg_name == "cube4D":
            selected_sub_obj = Cube4DRotations(N=N, dimensions=dimensions, **kwargs)
        elif alg_name == "zero":
            selected_sub_obj = ZeroRotations(N=N, dimensions=dimensions, **kwargs)
        elif alg_name == "ico":
            selected_sub_obj = IcoRotations(N=N, dimensions=dimensions, **kwargs)
        elif alg_name == "cube3D":
            selected_sub_obj = Cube3DRotations(N=N, dimensions=dimensions, **kwargs)
        else:
            raise ValueError(f"The algorithm {alg_name} not familiar to QuaternionGridFactory.")
        selected_sub_obj.gen_grid()
        return selected_sub_obj


class ConvergenceSphereGridFactory:

    """
    This is a central object for studying the convergence of SphereGridNDim objects with the number of dimensions.
    """

    def __init__(self, alg_name: str, dimensions: int, N_set = None, use_saved=True, **kwargs):
        if N_set is None:
            N_set = SMALL_NS
        self.N_set = N_set
        self.alg_name = alg_name
        self.dimensions = dimensions
        self.use_saved = use_saved
        self.kwargs = kwargs
        self.list_sphere_grids = []

    def get_name(self):
        N_min = self.N_set[0]
        N_max = self.N_set[-1]
        N_len = len(self.N_set)
        return f"convergence_{self.alg_name}_{self.dimensions}d_{N_min}_{N_max}_{N_len}"

    def create(self) -> list:
        list_sphere_grids = []
        for N in self.N_set:
            sg = SphereGridFactory.create(self.alg_name, N, self.dimensions, use_saved=self.use_saved, **self.kwargs)
            list_sphere_grids.append(sg)
        return list_sphere_grids

    @save_or_use_saved
    def get_list_sphere_grids(self):
        if not self.list_sphere_grids:
            self.list_sphere_grids = self.create()
        return self.list_sphere_grids

    @save_or_use_saved
    def get_spherical_voronoi_areas(self):
        """
        Get plotting-ready data on the size of Voronoi areas for different numbers of points.
        """
        data = []
        for N, sg in zip(self.N_set, self.get_list_sphere_grids()):
            real_N = len(sg.get_grid_as_array())
            ideal_area = 4*pi/real_N
            try:
                real_areas = sg.get_voronoi_areas()
            except ValueError:
                real_areas = [np.NaN] * real_N
            for area in real_areas:
                data.append([real_N, ideal_area, area])
        df = pd.DataFrame(data, columns=["N", "ideal area", "sph. Voronoi cell area"])
        return df

    @save_or_use_saved
    def get_generation_times(self, repeats=5):
        """
        Get plotting-ready data on time needed to generate spherical grids of different sizes. Each generation is
        repeated several times to be able to estimate the error.
        """
        data = []
        for N in self.N_set:
            for _ in range(repeats):
                t1 = time()
                # cannot use saved if you are timing!
                sg = SphereGridFactory.create(self.alg_name, N, self.dimensions, use_saved=False, **self.kwargs)
                t2 = time()
                data.append([len(sg.get_grid_as_array()), t2 - t1])
        df = pd.DataFrame(data, columns=["N", "Time [s]"])
        return df


if __name__ == "__main__":
    SphereGridFactory.create("cube4D", 200, 3, use_saved=False, print_messages=True, filter_non_unique=True)




"""
Abstract implementation of plots and multi-plots.

All other plots extend RepresentationCollection. Each RepresentationCollection has a set of methods used to view the
data of the object with which the Collection is initiated in different ways. Plotting methods always posses ax and fig
arguments so that several plots can be plotted on top of each other or next to each other.

MultiRepresentationCollection provides functionality for combining several plots in one image.
PanelRepresentationCollection specifically creates one sub-plot for each grid-generating algorithm.
"""

from abc import ABC
from typing import Union, List

import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.animation import FuncAnimation, PillowWriter, Animation
from matplotlib.pyplot import Figure
from matplotlib.axes import Axes
from mpl_toolkits.mplot3d.axes3d import Axes3D
from numpy.typing import NDArray
from IPython import display
from scipy.spatial import geometric_slerp

from molgri.constants import DIM_SQUARE, DEFAULT_DPI, EXTENSION_FIGURES, DEFAULT_DPI_MULTI, EXTENSION_ANIMATIONS
from molgri.logfiles import paths_free_4_all
from molgri.paths import PATH_OUTPUT_PLOTS, PATH_OUTPUT_ANIS
from molgri.space.utils import normalise_vectors


def _set_style_and_context(context: str = None, color_style: str = None):
    """
    This method should be run before an axis is created. Set up size of elements (paper, notebook, talk, poster),
    color scheme (white, dark ...) and similar. If nothing selected, use class defaults.

    Args:
        context: forwarded to sns.set_context -> should be one of: paper, notebook, talk, poster
        color_style: describes the general color scheme of the plots (dark, white)

    """
    sns.reset_orig()
    sns.set_context(context)
    plt.style.use('default')
    if "dark" in color_style:
        plt.style.use('dark_background')


class RepresentationCollection(ABC):
    """
    Most general plotting class. Set all methods that could be useful for many various types of plots here.
    All other plots, including multiplots, inherit from this class.
    """

    def __init__(self, data_name: str, fig_path: str = PATH_OUTPUT_PLOTS,
                 ani_path: str = PATH_OUTPUT_ANIS, default_figsize: tuple = DIM_SQUARE,
                 default_context: str = None, default_color_style: str = None,
                 default_complexity_level: str = None, default_axes_limits: tuple = (None,)*6):
        """
        In the __init__, defaults should be set that will generally be used for an entire class of figures - but it
        is still possible to not use defaults for some of the plots.

        Args:
            data_name: keyword that combines all figures of a specific sub-class
            fig_path: folder to save figures if created, should be set in subclasses
            ani_path: folder to save animations if created, should be set in subclasses
            default_context: forwarded to sns.set_context -> should be one of: paper, notebook, talk, poster
            default_color_style: describes the general color scheme of the plots (dark, white)
            default_complexity_level: select from full, half_empty, empty

        """
        # defaults
        if default_context is None:
            default_context = "notebook"
        if default_color_style is None:
            default_color_style = "white"
        if default_complexity_level is None:
            default_complexity_level = "full"

        self.data_name = data_name
        self.fig_path = fig_path
        self.ani_path = ani_path
        self.figsize = default_figsize
        self.context = default_context
        self.color_style = default_color_style
        self.complexity_level = default_complexity_level
        self.axes_limits = default_axes_limits

        # current values, will be overwritten by each new plot
        self.ax = None
        self.fig = None

    ##################################################################################################################
    #                                   SET-UP AXES & SAVE IN THE END
    ##################################################################################################################

    def _create_fig_ax(self, fig: Figure = None, ax: Union[Axes, Axes3D] = None, **creation_kwargs):
        """
        This function sets up a plot will create fig and ax in appropriate style. Each plotting function must call it.
        """
        context = creation_kwargs.pop("context", self.context)
        complexity = creation_kwargs.pop("complexity", self.complexity_level)
        color_style = creation_kwargs.pop("color_style", self.color_style)
        figsize = creation_kwargs.pop("figsize", self.figsize)
        projection = creation_kwargs.pop("projection", None)

        _set_style_and_context(context=context, color_style=color_style)
        if fig is None:
            self.fig = plt.figure(figsize=figsize)
        else:
            self.fig = fig
        if ax is None:
            self.ax = self.fig.add_subplot(111, projection=projection)
        else:
            self.ax = ax
        self.__set_complexity_level(complexity)
        self.__set_style_part_2(color_style=color_style)

    def _save_plot_type(self, plot_type_name: str, **saving_kwargs):
        save_ending = saving_kwargs.pop("save_ending", EXTENSION_FIGURES)
        dpi = saving_kwargs.pop("dpi", DEFAULT_DPI)
        path = paths_free_4_all([f"{self.data_name}_{plot_type_name}"], [save_ending], [self.fig_path])[0]
        self.fig.tight_layout()
        plt.savefig(path, dpi=dpi, bbox_inches='tight')
        plt.close()

    def _save_animation_type(self, animation: Animation, ani_type_name: str, fps: int = 10,
                             dpi: int = DEFAULT_DPI_MULTI):
        path = paths_free_4_all([f"{self.data_name}_{ani_type_name}"], [EXTENSION_ANIMATIONS], [self.ani_path])[0]
        writergif = PillowWriter(fps=fps, bitrate=-1)
        # noinspection PyTypeChecker
        animation.save(path, writer=writergif, dpi=dpi)

    def __set_complexity_level(self, complexity_level: str):
        if complexity_level == "empty" or complexity_level == "half_empty":
            self.ax.set_xticks([])
            self.ax.set_yticks([])
            if isinstance(self.ax, Axes3D):
                self.ax.set_zticks([])

            if complexity_level == "empty":
                self.ax.axis('off')

    def __set_style_part_2(self, color_style: str):
        color = (0.5, 0.5, 0.5, 0.7)
        if color_style == "half_dark":
            self.ax.xaxis.set_pane_color(color)
            self.ax.yaxis.set_pane_color(color)
            if isinstance(self.ax, Axes3D):
                self.ax.zaxis.set_pane_color(color)

    ##################################################################################################################
    #                             USEFUL FUNCTIONS TO BE CALLED BY SUB-CLASSES
    ##################################################################################################################

    def get_possible_title(self):
        return self.data_name

    def _equalize_axes(self):
        """
        Makes x, y, (z) axis equally longs and if limits given, enforces them on all axes.
        If you also use set_axis_limits, this should be run afterwards!
        """
        if isinstance(self.ax, Axes3D):
            # because ax.set_aspect('equal') does not work for 3D axes
            self.ax.set_box_aspect(aspect=[1, 1, 1])
            x_lim, y_lim, z_lim = self.ax.get_xlim3d(), self.ax.get_ylim3d(), self.ax.get_zlim3d()
            all_ranges = abs(x_lim[1] - x_lim[0]), abs(y_lim[1] - y_lim[0]), abs(z_lim[1] - z_lim[0])
            x_middle, y_middle, z_middle = np.mean(x_lim), np.mean(y_lim), np.mean(z_lim)
            plot_range = 0.5 * max(all_ranges)
            self.ax.set_xlim3d([x_middle - plot_range, x_middle + plot_range])
            self.ax.set_ylim3d([y_middle - plot_range, y_middle + plot_range])
            self.ax.set_zlim3d([z_middle - plot_range, z_middle + plot_range])
        else:
            self.ax.set_aspect('equal')

    def _set_axis_limits(self, limits: tuple = None):
        """
        List limits in a 4- or 6-element tuple in the order: x_min, x_max, y_min, y_max, z_min, z_max.
        If a plot is 2D, z values will be ignored. If any value should not be set, use None for that element.

        Args:
            limits: a tuple including all 4 or 6 limits
        """
        if limits is None:
            limits = self.axes_limits
        assert len(limits) == 4 or len(limits) == 6, "Wrong number of limit values"

        x_min_limit = limits[0]
        x_max_limit = limits[1]
        y_min_limit = limits[2]
        y_max_limit = limits[3]

        self.ax.set_xlim(x_min_limit, x_max_limit)
        self.ax.set_ylim(y_min_limit, y_max_limit)

        if isinstance(self.ax, Axes3D):
            z_min_limit = limits[4]
            z_max_limit = limits[5]
            self.ax.set_zlim(z_min_limit, z_max_limit)

    def _animate_figure_view(self, fig, ax, ani_name="rotated", **kwargs) -> FuncAnimation:
        """
        Call after you have created some 3D figure.
        Rotate the 3D figure for 360 degrees around itself and save the animation.
        """

        def animate(frame):
            # rotate the view left-right
            ax.view_init(azim=2*frame)
            return fig

        anim = FuncAnimation(fig, animate, frames=180, interval=50)
        dpi = kwargs.pop("dpi", 200)
        self._save_animation_type(anim, ani_name, fps=10, dpi=dpi)
        return anim


class MultiRepresentationCollection(ABC):

    def __init__(self, data_name, list_plots: List[RepresentationCollection], figsize=None,
                 n_rows: int = 4, n_columns: int = 2):
        self.n_rows = n_rows
        self.data_name = data_name
        self.n_columns = n_columns
        assert self.n_rows*self.n_columns == len(list_plots), f"{self.n_rows}x{self.n_columns}=/={len(list_plots)}"
        self.figsize = figsize
        if self.figsize is None:
            self.figsize = self.n_columns * DIM_SQUARE[0], self.n_rows * DIM_SQUARE[1]

        self.list_plots = list_plots

        # current values, to be overwritten with each plotting function
        self.fig = None
        self.all_ax = None

    ##################################################################################################################
    #                                   SET-UP AXES & SAVE IN THE END
    ##################################################################################################################

    def _make_plot_for_all(self, plotting_method: str, all_ax: NDArray = None, plotting_kwargs: dict = None,
                           remove_midlabels: bool = True, projection: str = None, creation_kwargs: dict = None):
        """This method should be run in all sub-modules to set-up the axes"""
        if plotting_kwargs is None:
            plotting_kwargs = dict()
        if creation_kwargs is None:
            creation_kwargs = dict()
        _set_style_and_context(context=self.list_plots[0].context, color_style=self.list_plots[0].color_style)
        self.__create_fig_ax(all_ax=all_ax, projection=projection, **creation_kwargs)
        for ax, subplot in zip(self.all_ax.ravel(), self.list_plots):
            plot_func = getattr(subplot, plotting_method)
            plot_func(fig=self.fig, ax=ax, save=False, **plotting_kwargs)
        if remove_midlabels:
            self.__remove_midlabels()

    def __create_fig_ax(self, all_ax=None, sharex="all", sharey="all", projection=None, figsize=None):
        if figsize is None:
            figsize = self.figsize
        if all_ax:
            self.fig = plt.figure(figsize=figsize)
            self.all_ax = all_ax
        else:
            self.fig, self.all_ax = plt.subplots(self.n_rows, self.n_columns, subplot_kw={'projection': projection},
                                                 sharex=sharex, sharey=sharey, figsize=figsize)

    def __remove_midlabels(self):
        if not isinstance(self.all_ax[0], Axes3D):
            if self.n_rows > 1 and self.n_columns > 1:
                for my_a in self.all_ax[:, 1:]:
                    for subax in my_a:
                        subax.set_ylabel("")
                for my_a in self.all_ax[:-1, :]:
                    for subax in my_a:
                        subax.set_xlabel("")
            else:
                for my_a in self.all_ax[1:]:
                    my_a.set_ylabel("")

    def _save_multiplot(self, multiplot_type, save_ending: str = EXTENSION_FIGURES, dpi: int = DEFAULT_DPI, **kwargs):
        self.fig.tight_layout()
        fig_path = self.list_plots[0].fig_path
        self.fig.savefig(f"{fig_path}{self.data_name}_{multiplot_type}.{save_ending}", dpi=dpi,
                         bbox_inches='tight', **kwargs)
        plt.close()

    def save_multianimation(self, animation, plot_type, dpi=100, fps=10):
        writergif = PillowWriter(fps=fps, bitrate=-1)
        ani_path = self.list_plots[0].ani_path
        # noinspection PyTypeChecker
        animation.save(f"{ani_path}{self.data_name}_{plot_type}.gif", writer=writergif, dpi=dpi)

    ##################################################################################################################
    #                             USEFUL FUNCTIONS TO BE CALLED BY SUB-CLASSES
    ##################################################################################################################

    def add_titles(self, list_titles, **title_kwargs):
        for title, ax in zip(list_titles, self.all_ax.ravel()):
            ax.set_title(title, **title_kwargs)

    def set_log_scale(self, x_axis=True, y_axis=True):
        for ax in self.all_ax.ravel():
            if x_axis:
                ax.set_xscale("log")
            if y_axis:
                ax.set_yscale("log")

    def add_legend(self, **legend_kwargs):
        """
        This adds one overall legend to the entire multiplot, only using unique handles.
        """
        all_handles = []
        all_labels = []
        for ax in self.all_ax.ravel():
            handles, labels = ax.get_legend_handles_labels()
            all_labels.extend(labels)
            all_handles.extend(handles)
        # remove duplicates
        unique = [(h, l) for i, (h, l) in enumerate(zip(all_handles, all_labels)) if l not in all_labels[:i]]
        self.fig.legend(*zip(*unique), **legend_kwargs)

    def unify_axis_limits(self, x_ax=True, y_ax=True, z_ax=True):
        """
        Unifies the min and max ranges of all subplots for selected axes.
        """
        x_lim = [np.infty, -np.infty]
        y_lim = [np.infty, -np.infty]
        z_lim = [np.infty, -np.infty]
        # determine the max and min values
        for i, subaxis in enumerate(self.all_ax.ravel()):
            if subaxis.get_xlim()[0] < x_lim[0]:
                x_lim[0] = subaxis.get_xlim()[0]
            if subaxis.get_xlim()[1] > x_lim[1]:
                x_lim[1] = subaxis.get_xlim()[1]
            if subaxis.get_ylim()[0] < y_lim[0]:
                y_lim[0] = subaxis.get_ylim()[0]
            if subaxis.get_ylim()[1] > y_lim[1]:
                y_lim[1] = subaxis.get_ylim()[1]
            if isinstance(subaxis, Axes3D) and subaxis.get_zlim()[0] < z_lim[0]:
                z_lim[0] = subaxis.get_zlim()[0]
            if isinstance(subaxis, Axes3D) and subaxis.get_zlim()[1] > z_lim[1]:
                z_lim[1] = subaxis.get_zlim()[1]
        # change them all
        for i, subaxis in enumerate(self.all_ax.ravel()):
            if x_ax:
                subaxis.set_xlim(*x_lim)
            if y_ax:
                subaxis.set_ylim(*y_lim)
            if isinstance(subaxis, Axes3D) and z_ax:
                subaxis.set_zlim(*z_lim)

    def animate_figure_view(self, plot_type: str, dpi=100) -> FuncAnimation:
        """
        Rotate all 3D figures for 360 degrees around themselves and save the animation.
        """

        def animate(frame):
            # rotate the view left-right
            for ax in self.all_ax.ravel():
                ax.view_init(azim=2*frame)
            return self.fig

        anim = FuncAnimation(self.fig, animate, frames=180, interval=50)
        self.save_multianimation(anim, plot_type=plot_type, dpi=dpi)
        return anim

    def add_colorbar(self, **cbar_kwargs):
        orientation = cbar_kwargs.pop("orientation", "vertical")
        cbar_label = cbar_kwargs.pop("cbar_label", None)
        all_collections = []
        for ax in self.all_ax.ravel():
            # this doesn't overwrite, but adds points to the collection!
            all_collections = ax.collections[0]

        cbar = self.fig.colorbar(all_collections, ax=self.all_ax[-1])
        # if orientation == "horizontal":
        #     pos = self.all_ax.ravel()[-1].get_position()
        #     cax = self.fig.add_axes([pos.x0 -0.05, pos.y0 -0.1, 0.3, 0.03])
        #
        # else:
        #     pos = self.all_ax.ravel()[-1].get_position()
        #     cax = self.fig.add_axes([pos.x1 - 1, pos.y0 + 0.0, 0.03, 0.4])
        # cbar = self.fig.colorbar(all_collections, cax=cax, orientation=orientation)
        # if orientation == "vertical":
        #     cax.yaxis.set_label_position('left')
        # if cbar_label:
        #     cbar.set_label(cbar_label)


class PanelRepresentationCollection(MultiRepresentationCollection):

    def __init__(self, data_name, list_plots: List[RepresentationCollection], landscape=True, figsize=None):
        if landscape:
            n_rows = 2
            n_columns = 3
        else:
            n_rows = 3
            n_columns = 2
        super().__init__(data_name, list_plots, n_columns=n_columns, n_rows=n_rows, figsize=figsize)


def show_anim_in_jupyter(anim):
    video = anim.to_html5_video()
    html = display.HTML(video)
    display.display(html)
    # necessary to not create an extra empty window
    plt.close()


def plot_voronoi_cells(sv, ax, plot_vertex_points=True):
    sv.sort_vertices_of_regions()
    t_vals = np.linspace(0, 1, 2000)
    # plot Voronoi vertices
    if plot_vertex_points:
        ax.scatter(sv.vertices[:, 0], sv.vertices[:, 1], sv.vertices[:, 2], c='g')
    # indicate Voronoi regions (as Euclidean polygons)
    for region in sv.regions:
        n = len(region)
        for j in range(n):
            start = sv.vertices[region][j]
            end = sv.vertices[region][(j + 1) % n]
            norm = np.linalg.norm(start)
            result = geometric_slerp(normalise_vectors(start), normalise_vectors(end), t_vals)
            ax.plot(norm * result[..., 0], norm * result[..., 1], norm * result[..., 2], c='k')



"""
Combine a ParsedTrajectory and a FullGrid to generate a MSM or SqRA model.

In this module, the two methods of evaluating transitions between states - the MSM and the SqRA approach - are
implemented.
"""
from abc import ABC, abstractmethod
from typing import Tuple, Sequence, Any

import numpy as np
from numpy.typing import NDArray
from scipy.sparse.linalg import eigs
from scipy.constants import k as kB
from tqdm import tqdm

from molgri.wrappers import save_or_use_saved
from molgri.molecules.parsers import ParsedTrajectory
from molgri.space.fullgrid import FullGrid


class SimulationHistogram:

    """
    This class takes the data from ParsedTrajectory and combines it with a specific FullGrid so that each simulation
    step can be assigned to a cell and the occupancy of those cells evaluated.
    """

    def __init__(self, parsed_trajectory: ParsedTrajectory, full_grid: FullGrid):
        self.parsed_trajectory = parsed_trajectory
        self.full_grid = full_grid

    # noinspection PyMissingOrEmptyDocstring
    def get_name(self) -> str:
        traj_name = self.parsed_trajectory.get_name()
        grid_name = self.full_grid.get_name()
        return f"{traj_name}_{grid_name}"

    def get_all_assignments(self) -> Tuple[NDArray, NDArray]:
        """
        For each step in the trajectory assign which cell it belongs to. Uses the default atom selection of the
        ParsedTrajectory object.

        Returns:
            (all centers of mass within grid, all assignments to grid cells)
        """
        atom_selection = self.parsed_trajectory.default_atom_selection
        # if you do nan_free, your tau may not be correct anymore because several steps in-between may be missing
        return self.parsed_trajectory.assign_coms_2_grid_points(self.full_grid, atom_selection=atom_selection,
                                                                nan_free=False)


class TransitionModel(ABC):

    """
    A class that contains both the MSM and SQRA models.
    """

    def __init__(self, sim_hist: SimulationHistogram, tau_array: NDArray = None, use_saved: bool = False):
        self.sim_hist = sim_hist
        if tau_array is None:
            tau_array = np.array([2, 5, 7, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 110, 130, 150, 180, 200])
        self.tau_array = tau_array
        self.use_saved = use_saved
        _, self.assignments = self.sim_hist.get_all_assignments()
        self.num_cells = len(self.sim_hist.full_grid.get_flat_position_grid())
        self.num_tau = len(self.tau_array)
        self.transition_matrix = None

    # noinspection PyMissingOrEmptyDocstring
    def get_name(self) -> str:
        return self.sim_hist.get_name()

    @abstractmethod
    def get_transitions_matrix(self) -> NDArray:
        """For MSM, generate the transition matrix from simulation data. For SQRA, generate rate matrix from the
        point energy calculations.

        Returns:
            an array of shape (num_tau, num_cells, num_cells) for MSM or (n1, num_cells, num_cells) for SQRA
        """
        pass

    @save_or_use_saved
    def get_eigenval_eigenvec(self, num_eigenv: int = 15, **kwargs) -> Tuple[NDArray, NDArray]:
        """
        Obtain eigenvectors and eigenvalues of the transition matrices.

        Args:
            num_eigenv: how many eigenvalues/vectors pairs to return (too many may give inaccurate results)
            **kwargs: named arguments to forward to eigs()
        Returns:
            (eigenval, eigenvec) a tuple of eigenvalues and eigenvectors, first num_eigv given for all tau-s
            Eigenval is of shape (num_tau, num_eigenv), eigenvec of shape (num_tau, num_cells, num_eigenv)
        """
        all_tms = self.get_transitions_matrix()
        all_eigenval = np.zeros((self.num_tau, num_eigenv))
        all_eigenvec = np.zeros((self.num_tau, self.num_cells, num_eigenv))
        for tau_i, tau in enumerate(self.tau_array):
            tm = all_tms[tau_i]  # the transition matrix for this tau
            tm[np.isnan(tm)] = 0  # replace nans with zeros
            # in order to compute left eigenvectors, compute right eigenvectors of the transpose
            eigenval, eigenvec = eigs(tm.T, num_eigenv, maxiter=100000, tol=0, which="LM", sigma=0, **kwargs)
            # don't need to deal with complex outputs in case all values are real
            # TODO: what happens here if we have negative imaginary components?
            if eigenvec.imag.max() == 0 and eigenval.imag.max() == 0:
                eigenvec = eigenvec.real
                eigenval = eigenval.real
            # sort eigenvectors according to their eigenvalues
            idx = eigenval.argsort()[::-1]
            eigenval = eigenval[idx]
            eigenvec = eigenvec[:, idx]
            all_eigenval[tau_i] = eigenval
            all_eigenvec[tau_i] = eigenvec
        return all_eigenval, all_eigenvec


class MSM(TransitionModel):

    """
    Markov state model (MSM) works on simulation trajectories by discretising the accessed space and counting
    transitions between states in the time span of tau.
    """

    @save_or_use_saved
    def get_transitions_matrix(self, noncorr: bool = False) -> NDArray:
        """
        Obtain a set of transition matrices for different tau-s specified in self.tau_array.

        Args:
            noncorr: bool, should only every tau-th frame be used for MSM construction
                     (if False, use sliding window - much more expensive but throws away less data)
        Returns:
            an array of transition matrices of shape (self.num_tau, self.num_cells, self.num_cells)
        """

        def window(seq: Sequence, len_window: int, step: int = 1) -> Tuple[Any, Any]:
            """
            How this works: returns a list of sublists. Each sublist has two elements: [start_element, end_element].
            Both elements are taken from the seq and are len_window elements apart. The parameter step controls what
            the step between subsequent start_elements is.

            Yields:
                [start_element, end_element] where both elements com from seq where they are len_window positions
                apart

            Example:
                >>> gen_window = window([1, 2, 3, 4, 5, 6], len_window=2, step=3)
                >>> next(gen_window)
                (1, 3)
                >>> next(gen_window)
                (4, 6)
            """
            # in this case always move the window by step and use all points in simulations to count transitions
            for k in range(0, len(seq) - len_window, step):
                start_stop_list = seq[k: k + len_window + 1:len_window]
                if not np.isnan(start_stop_list).any():
                    yield tuple([int(el) for el in start_stop_list])

        def noncorr_window(seq: Sequence, len_window: int) -> Tuple[Any, Any]:
            """
            Subsample the seq so that only each len_window-th element remains and then similarly return pairs of
            elements.

            Example:
                >>> gen_obj = noncorr_window([1, 2, 3, 4, 5, 6, 7], 3)
                >>> next(gen_obj)
                (1, 4)
                >>> next(gen_obj)
                (4, 7)

            """
            # in this case, only use every len_window-th element for MSM. Faster but loses a lot of data
            return window(seq, len_window, step=len_window)

        if self.transition_matrix is None:
            self.transition_matrix = np.zeros(shape=(self.num_tau, self.num_cells, self.num_cells))
            for tau_i, tau in enumerate(tqdm(self.tau_array)):
                # save the number of transitions between cell with index i and cell with index j
                count_per_cell = {(i, j): 0 for i in range(self.num_cells) for j in range(self.num_cells)}
                if not noncorr:
                    window_cell = window(self.assignments, int(tau))
                else:
                    window_cell = noncorr_window(self.assignments, int(tau))
                for cell_slice in window_cell:
                    try:
                        count_per_cell[cell_slice] += 1
                    except KeyError:
                        # the point is outside the grid and assigned to NaN - ignore for now
                        pass
                for key, value in count_per_cell.items():
                    start_cell, end_cell = key
                    self.transition_matrix[tau_i, start_cell, end_cell] += value
                    # enforce detailed balance
                    self.transition_matrix[tau_i, end_cell, start_cell] += value
                # divide each row of each matrix by the sum of that row
                sums = self.transition_matrix[tau_i].sum(axis=-1, keepdims=True)
                sums[sums == 0] = 1
                self.transition_matrix[tau_i] = self.transition_matrix[tau_i] / sums
        return self.transition_matrix


class SQRA(TransitionModel):

    """
    As opposed to MSM, this object works with a pseudo-trajectory that evaluates energy at each grid point and
    with geometric parameters of position space division.
    """

    def __init__(self, sim_hist: SimulationHistogram, energy_type: str = "Potential", **kwargs):
        # SQRA doesn't need a tau array, but for compatibility with MSM we use this one
        tau_array = np.array([1])
        self.energy_type = energy_type
        super().__init__(sim_hist, tau_array=tau_array, **kwargs)

    @save_or_use_saved
    def get_transitions_matrix(self, D: float = 1, T=273) -> NDArray:
        """
        Return the rate matrix as calculated by the SqRA formula:

        Q_ij = np.sqrt(pi_j/pi_i) * D * S_ij / (h_ij * V_i)

        Args:
            D: diffusion constant
            energy_type: the keyword to pass to self.sim_hist.parsed_trajectory.get_all_energies to obtain energy
            information at centers of cells

        Returns:
            a (1, self.num_cells, self.num_cells) array that is a rate matrix estimated with SqRA (first dimension
            is expanded to be compatible with the MSM model)
        """
        if self.transition_matrix is None:
            voronoi_grid = self.sim_hist.full_grid.get_full_voronoi_grid()
            all_volumes = voronoi_grid.get_all_voronoi_volumes()
            # TODO: move your work to sparse matrices at some point?
            all_surfaces = voronoi_grid.get_all_voronoi_surfaces_as_numpy()
            all_distances = voronoi_grid.get_all_distances_between_centers_as_numpy()
            # energies are either NaN (calculation in that cell was not completed or the particle left the cell during
            # optimisation) or they are the arithmetic average of all energies of COM assigned to that cell.
            all_energies = np.empty(shape=(self.num_cells,))
            energy_counts = np.zeros(shape=(self.num_cells,))
            obtained_energies = self.sim_hist.parsed_trajectory.get_all_energies(energy_type=self.energy_type)
            for a, e in zip(self.assignments, obtained_energies):
                if not np.isnan(a):
                    all_energies[int(a)] += e
                    energy_counts[int(a)] += 1
            # in both cases avoiding division with zero
            # TODO: instead of averaging find the most central point
            all_energies = np.divide(all_energies, energy_counts, out=np.zeros_like(all_energies),
                                     where=energy_counts != 0)
            self.transition_matrix = np.divide(D * all_surfaces, all_distances, out=np.zeros_like(D * all_surfaces),
                                    where=all_distances!=0)

            for i, _ in enumerate(self.transition_matrix):
                divide_by = all_volumes[i]
                self.transition_matrix[i] = np.divide(self.transition_matrix[i], divide_by,
                                                      out=np.zeros_like(self.transition_matrix[i]),
                                         where=(divide_by != 0 and divide_by != np.NaN))
            for j, _ in enumerate(self.transition_matrix):
                for i, _ in enumerate(self.transition_matrix):
                    # gromacs uses kJ/mol as energy unit, boltzmann constant is J/K
                    multiply_with = np.exp((all_energies[i]-all_energies[j])*1000/(2*kB*T))
                    self.transition_matrix[i, j] = np.multiply(self.transition_matrix[i, j], multiply_with,
                                                               out=np.zeros_like(self.transition_matrix[i, j]),
                                               where=multiply_with != np.NaN)

            # normalise rows
            sums = np.sum(self.transition_matrix, axis=1)
            np.fill_diagonal(self.transition_matrix, -sums)
            # additional axis
            self.transition_matrix = self.transition_matrix[np.newaxis, :]
            assert self.transition_matrix.shape == (1, self.num_cells, self.num_cells)
        return self.transition_matrix


if __name__ == "__main__":
    import pandas as pd
    from molgri.molecules.parsers import FileParser, ParsedEnergy

    # preparing the parsed trajectory
    path_energy = "/home/mdglasius/Modelling/trypsin_normal/nobackup/outputs/data.txt"
    df = pd.read_csv(path_energy)
    energies = df['Potential Energy (kJ/mole)'].to_numpy()[:, np.newaxis]
    pe = ParsedEnergy(energies=energies, labels=["Potential Energy"], unit="(kJ/mole)")
    pt_parser = FileParser(
        path_topology="/home/mdglasius/Modelling/trypsin_normal/inputs/trypsin_probe.pdb",
        path_trajectory="/home/mdglasius/Modelling/trypsin_normal/nobackup/outputs/aligned_traj.dcd")
    parsed_trajectory = pt_parser.get_parsed_trajectory(default_atom_selection="segid B")
    parsed_trajectory.energies = pe

    # preparing the grid
    fg = FullGrid(t_grid_name="[5, 10, 15]", o_grid_name="ico_100", b_grid_name="zero")

    sh = SimulationHistogram(parsed_trajectory, fg)
    my_msm = MSM(sh, tau_array=np.array([10, 50, 100]), use_saved=False)



"""
All project-applicable wrappers.

Specifically, functionality for:
 - timing the function and writing out the result
 - using saved data if it exists
 - marking deprecated functions
"""

import os
import warnings
from functools import wraps
from time import time
from datetime import timedelta
import numpy as np
import pandas as pd
import pickle

from molgri.paths import OUTPUT_PLOTTING_DATA


def time_method(my_method):
    """
    This wrapper times the execution of a method and prints the duration of the execution.

    Args:
        my_method: method with first argument self, the class it belongs to must implement a method
                   self.get_decorator_name() label with value like 'grid ico_15' or
                   'pseudotrajectory H2O_HCl_ico_15_full'

    Returns:
        what my_method returns
    """
    @wraps(my_method)
    def decorated(*args, **kwargs):
        self_arg = args[0]

        t1 = time()
        func_value = my_method(*args, **kwargs)
        t2 = time()
        print(f"Timing the generation of {self_arg.get_decorator_name()}: ", end="")
        print(f"{timedelta(seconds=t2-t1)} hours:minutes:seconds")
        return func_value
    return decorated


def save_or_use_saved(my_method):
    """
    This can be used on any method that provides data for plotting (important if data production takes a long time).
    Able to save any python objects: if numpy array will save as .npy, if pandas DataFrame as .csv, everything else
    as pickle.

    Requirements:
        the class must have an attribute self.use_saved (bool)
        the class must have a method get_name() which provides a name that is suitable for saving (no whitespace etc)
        the method should not have any parameters that are commonly changed or their values should be part of the name
        (otherwise you risk reading and using data with old parameters)

    Args:
        my_method: around which method the wrapper is applied

    Returns:
        whatever the original method would return, either freshly created or read from a file
    """
    @wraps(my_method)
    def decorated(self, *args, **kwargs):
        method_name = my_method.__name__
        name_without_ext = f"{OUTPUT_PLOTTING_DATA}{method_name}_{self.get_name()}"
        # try to find a suitable saved file
        if self.use_saved:
            if os.path.isfile(f"{name_without_ext}.npy"):
                return np.load(f"{name_without_ext}.npy")
            elif os.path.isfile(f"{name_without_ext}.csv"):
                return pd.read_csv(f"{name_without_ext}.csv", index_col=0)
            elif os.path.isfile(name_without_ext):
                with open(name_without_ext, 'rb') as f:
                    loaded_data = pickle.load(f)
                return loaded_data
            # else will simply continue
        # don't use else - the rest should be run if 1) not self.use_saved OR 2) file doesn't exist
        method_output = my_method(self, *args, **kwargs)
        if isinstance(method_output, pd.DataFrame):
            method_output.to_csv(f"{name_without_ext}.csv", index=True)
        elif isinstance(method_output, np.ndarray):
            np.save(f"{name_without_ext}.npy", method_output)
        else:
            with open(name_without_ext, 'wb') as f:
                pickle.dump(method_output, f)
        return method_output
    return decorated


def deprecated(my_method):
    """
    Mark functions or methods that may be removed in the future. Raises a warning.
    """
    warnings.warn(f"The method {my_method.__name__} is deprecated and may be removed in the future.",
                  DeprecationWarning, stacklevel=2)

    @wraps(my_method)
    def decorated(*args, **kwargs):
        return my_method(*args, **kwargs)

    return decorated



"""
Analysis of uniformity of 3D and 4D sphere grids.

Always performed by selecting random vectors and evaluating how many grid points fall within the spherical cap area
around this vector. Ideally, the ratio of points within the spherical cap and all points should be equal to the
ratio of sphere cap area and full sphere area. For quaternions, double-coverage must be considered.
"""
from typing import Tuple

import numpy as np
import pandas as pd
from numpy.typing import NDArray
from scipy.constants import pi

from molgri.space.utils import angle_between_vectors, random_sphere_points, random_quaternions, \
    randomise_quaternion_set_signs
from molgri.constants import DEFAULT_ALPHAS_3D, DEFAULT_ALPHAS_4D, TEXT_ALPHAS_3D, TEXT_ALPHAS_4D


def vector_within_alpha(central_vec: NDArray, side_vec: NDArray, alpha: float) -> NDArray[bool]:
    """
    Calculate the angle between each pair of vectors in zip(central_vec, side_vector). Return an array in which True
    is given where the angle is smaller than alpha and False where not.

    Args:
        central_vec: a single vector of shape (d,) or an array of vectors of shape (n1, d)
        side_vec: a single vector of shape (d,) or an array of vectors of shape (n2, d)
        alpha: in radians, angle around central_vec where we check if side_vectors occur

    Returns:
        array of bools of shape (n1, n2) if both vectors are 2D
                       of shape (n2,) if central_vec is 1D and side_vec 2D
                       of shape (n1,) if central_vec is 2D and side_vec 1D
                       of shape (1,) if both vectors are 1D
    """
    return angle_between_vectors(central_vec, side_vec) < alpha


def count_points_within_alpha(array_points: NDArray, central_vec: NDArray, alpha: float) -> int:
    """
    Count how many points fall within spherical cap with central angle alpha around the central vector.

    Args:
        array_points: each row is a point on a sphere that may or may not be within a spherical cap area
        central_vec: defines the center of spherical cap
        alpha: defines the width of spherical cap

    Returns:
        the number of rows within array_points that are vectors with endings in the spherical cap area.
    """
    # since a list of True or False is returned, the sum gives the total number of True elements
    return int(np.sum(vector_within_alpha(central_vec, array_points, alpha)))


def random_axes_count_points(array_points: NDArray, alpha: float, num_random_points: int = 1000) -> NDArray:
    """
    Repeat count_points_within_alpha for many random selections of the central vector. Using this, we get a good
    estimate of uniformity across the entire sphere.

    Args:
        array_points: each row is a point on a sphere
        alpha: defines the width of spherical cap
        num_random_points: how many random central vectors to test

    Returns:
        Array of shape (num_random_points,) where for each random central vector the ratio
        array points within spherical cal/all array points is saved
    """
    central_vectors = random_sphere_points(num_random_points)
    return _repeat_count(array_points, central_vectors, num_random_points, alpha)


def random_quaternions_count_points(array_points: np.ndarray, alpha: float, num_random_points: int = 1000):
    """
    Same as random_axes_count_points, but because we are dealing with quaternions, random axes are defined differently
    and quaternion signs are randomised first.
    """
    central_vectors = random_quaternions(num_random_points)
    central_vectors = randomise_quaternion_set_signs(central_vectors)
    array_points = randomise_quaternion_set_signs(array_points)
    return _repeat_count(array_points, central_vectors, num_random_points, alpha)


def _repeat_count(array_points, central_vectors, num_random_points, alpha):
    # common function for calculation ratios for vectors/quaternions
    all_ratios = np.zeros(num_random_points)
    for i, central_vector in enumerate(central_vectors):
        num_within = count_points_within_alpha(array_points, central_vector, alpha)
        all_ratios[i] = num_within/len(array_points)
    return all_ratios


def prepare_statistics(point_array: NDArray, alphas: list, d: int = 3, num_rand_points: int = 1000) -> Tuple:
    """
    Prepare statistics data that is used for ViolinPlots, either based on 3D points or on quaternions.

    Args:
        point_array: an array of points for which coverage is tested, must be of the shape (N, d)
        alphas: a list of angles in radians at which convergence is tested
        d: number of dimensions (3 or 4)
        num_rand_points: number of random axes around which convergence is tested

    Returns:
        (data_frame_summary, data_frame_full)
    """
    if alphas is None:
        if d == 3:
            alphas = DEFAULT_ALPHAS_3D
        else:
            alphas = DEFAULT_ALPHAS_4D
    assert d == 3 or d == 4, f"Statistics available only for 3 or 4 dimensions, not d={d}"
    assert point_array.shape[1] == d, f"Provided array not in shape (N, {d})"
    if d == 3:
        sphere_surface = 4 * pi

        def cone_area_f(a):
            """Cone area 3D sphere, depending on angle a"""
            return 2 * pi * (1 - np.cos(a))

        actual_coverages_f = random_axes_count_points
    else:
        # explanation: see https://scialert.net/fulltext/?doi=ajms.2011.66.70&org=11
        sphere_surface = 2 * pi ** 2  # full 4D sphere has area 2pi^2 r^3

        def cone_area_f(a):
            """Cone area 4D sphere, depending on angle a"""
            return 1 / 2 * sphere_surface * (2 * a - np.sin(2 * a)) / np.pi

        actual_coverages_f = random_quaternions_count_points
    # write out short version ("N points", "min", "max", "average", "SD"
    columns = ["alphas", "ideal coverages", "min coverage", "avg coverage", "max coverage", "standard deviation"]
    ratios_columns = ["coverages", "alphas", "ideal coverage"]
    ratios = [[], [], []]
    data = np.zeros((len(alphas), 6))  # 5 data columns for: alpha, ideal coverage, min, max, average, sd
    for i, alpha in enumerate(alphas):
        cone_area = cone_area_f(alpha)
        ideal_coverage = cone_area / sphere_surface
        actual_coverages = actual_coverages_f(point_array, alpha, num_random_points=num_rand_points)
        ratios[0].extend(actual_coverages)
        ratios[1].extend([alpha] * num_rand_points)
        ratios[2].extend([ideal_coverage] * num_rand_points)
        data[i][0] = alpha
        data[i][1] = ideal_coverage
        data[i][2] = np.min(actual_coverages)
        data[i][3] = np.average(actual_coverages)
        data[i][4] = np.max(actual_coverages)
        data[i][5] = np.std(actual_coverages)
    alpha_df = pd.DataFrame(data=data, columns=columns)
    alpha_df = alpha_df.set_index("alphas")
    ratios_df = pd.DataFrame(data=np.array(ratios).T, columns=ratios_columns)
    return alpha_df, ratios_df


def write_statistics(stat_data: pd.DataFrame, full_data: pd.DataFrame,
                     path_summary: str, path_full_data: str,
                     num_random: int, name: str, dimensions: int, print_message: bool = False):
    """
    The top-level method that writes full statistics and a summary to files. Only deals with messages/files, all
    calculations are outsourced to other functions

    Args:
        stat_data: summary data returned by prepare_statistics function
        full_data: full data returned by prepare_statistics function
        path_summary: where to save summary
        path_full_data: where to save full data
        num_random: number of random orientations when calculating statistics
        name: ID of the system
        dimensions: 3 or 4 dimensions (determine sphere and spherical area formulas)
        print_message: True if a summary message should be printed out
    """
    # first message (what measure you are using)
    newline = "\n"
    if dimensions == 3:
        m1 = f"STATISTICS: Testing the coverage of grid {name} using {num_random} " \
             f"random points on a sphere. This statistics is useful for the position grid."
        labels = TEXT_ALPHAS_3D
    elif dimensions == 4:
        m1 = f"STATISTICS: Testing the coverage of quaternions {name} using {num_random} " \
             f"random quaternions. This statistics is useful for the relative orientations."
        labels = TEXT_ALPHAS_4D
    else:
        raise ValueError("Dimensions must be 3 or 4.")
    m2 = f"We select {num_random} random axes and count the number of grid points that fall within the angle" \
         f"alpha (selected from {labels}) of this axis. For an" \
         f"ideally uniform grid, we expect the ratio of num_within_alpha/total_num_points to equal the ratio" \
         f"area_of_alpha_spherical_cap/area_of_sphere, which we call ideal coverage."
    if print_message:
        print(m1)
        print(stat_data)
    # dealing with the file
    with open(path_summary, "w") as f:
        f.writelines([m1, newline, newline, m2, newline, newline])
    stat_data.to_csv(path_summary, mode="a")
    full_data.to_csv(path_full_data, mode="w")



"""
Loading currently important examples locally.
"""
import pandas as pd
import numpy as np

from molgri.molecules.parsers import ParsedEnergy, FileParser, ParsedTrajectory


def load_simulation_data() -> ParsedTrajectory:
    path_energy = "/home/mdglasius/Modelling/trypsin_normal/nobackup/outputs/data.txt"
    df = pd.read_csv(path_energy)
    energies = df['Potential Energy (kJ/mole)'].to_numpy()[:, np.newaxis]
    pe = ParsedEnergy(energies=energies, labels=["Potential Energy"], unit="(kJ/mole)")
    pt_parser = FileParser(
             path_topology="/home/mdglasius/Modelling/trypsin_normal/inputs/trypsin_probe.pdb",
             path_trajectory="/home/mdglasius/Modelling/trypsin_normal/nobackup/outputs/aligned_traj.dcd")
    parsed_trajectory = pt_parser.get_parsed_trajectory(default_atom_selection="segid B")
    parsed_trajectory.energies = pe
    return parsed_trajectory


def load_molgri_data() -> ParsedTrajectory:
    path_energy = "/home/mdglasius/Modelling/trypsin_test/output/measurements/temp_minimized.csv"
    df = pd.read_csv(path_energy)
    energies = df['potential'].to_numpy()[:, np.newaxis]
    pe = ParsedEnergy(energies=energies, labels=["Potential Energy"], unit="(kJ/mole)")
    pt_parser = FileParser(
             path_topology="/home/mdglasius/Modelling/trypsin_test/output/pt_files/final_trypsin_NH4_o_ico_512_b_zero_1_t_2153868773.gro",
             path_trajectory="/home/mdglasius/Modelling/trypsin_test/output/pt_files/minimizedPT.pdb")
    parsed_trajectory = pt_parser.get_parsed_trajectory(default_atom_selection="not protein")
    parsed_trajectory.energies = pe
    return parsed_trajectory